"""TweetyNet model"""
import torch
from torch import nn
from torch.nn import functional as F


class Conv2dTF(nn.Conv2d):
    """Conv2d with padding behavior from Tensorflow

    adapted from
    https://github.com/mlperf/inference/blob/16a5661eea8f0545e04c86029362e22113c2ec09/others/edge/object_detection/ssd_mobilenet/pytorch/utils.py#L40
    as referenced in this issue:
    https://github.com/pytorch/pytorch/issues/3867#issuecomment-507025011

    used to maintain behavior of original implementation of TweetyNet that used Tensorflow 1.0 low-level API
    """
    def __init__(self, *args, **kwargs):
        super(Conv2dTF, self).__init__(*args, **kwargs)
        self.padding = kwargs.get("padding", "SAME")

    def _compute_padding(self, input, dim):
        input_size = input.size(dim + 2)
        filter_size = self.weight.size(dim + 2)
        effective_filter_size = (filter_size - 1) * self.dilation[dim] + 1
        out_size = (input_size + self.stride[dim] - 1) // self.stride[dim]
        total_padding = max(
            0, (out_size - 1) * self.stride[dim] + effective_filter_size - input_size
        )
        additional_padding = int(total_padding % 2 != 0)

        return additional_padding, total_padding

    def forward(self, input):
        if self.padding == "VALID":
            return F.conv2d(
                input,
                self.weight,
                self.bias,
                self.stride,
                padding=0,
                dilation=self.dilation,
                groups=self.groups,
            )
        rows_odd, padding_rows = self._compute_padding(input, dim=0)
        cols_odd, padding_cols = self._compute_padding(input, dim=1)
        if rows_odd or cols_odd:
            input = F.pad(input, [0, cols_odd, 0, rows_odd])

        return F.conv2d(
            input,
            self.weight,
            self.bias,
            self.stride,
            padding=(padding_rows // 2, padding_cols // 2),
            dilation=self.dilation,
            groups=self.groups,
        )


class TweetyNet(nn.Module):
    def __init__(self,
                 num_classes,
                 input_shape=(1, 513, 88),
                 conv1_filters=32,
                 conv1_kernel_size=(5, 5),
                 conv2_filters=64,
                 conv2_kernel_size=(5, 5),
                 pool1_size=(8, 1),
                 pool1_stride=(8, 1),
                 pool2_size=(8, 1),
                 pool2_stride=(8, 1),
                 ):
        """initialize TweetyNet model

        Parameters
        ----------
        num_classes : int
            number of classes to predict, e.g., number of syllable classes in an individual bird's song
        input_shape : tuple
            with 3 elements corresponding to dimensions of spectrogram windows: (channels, frequency bins, time bins).
            i.e. we assume input is a spectrogram and treat it like an image, typically with one channel,
            the rows are frequency bins, and the columns are time bins. Default is (1, 513, 88).
        conv1_filters : int
            Number of filters in first convolutional layer. Default is 32.
        conv1_kernel_size : tuple
            Size of kernels, i.e. filters, in first convolutional layer. Default is (5, 5).
        conv2_filters : int
            Number of filters in second convolutional layer. Default is 64.
        conv2_kernel_size : tuple
            Size of kernels, i.e. filters, in second convolutional layer. Default is (5, 5).
        pool1_size : two element tuple of ints
            Size of sliding window for first max pooling layer. Default is (1, 8)
        pool1_stride : two element tuple of ints
            Step size for sliding window of first max pooling layer. Default is (1, 8)
        pool2_size : two element tuple of ints
            Size of sliding window for second max pooling layer. Default is (1, 8),
        pool2_stride : two element tuple of ints
            Step size for sliding window of second max pooling layer. Default is (1, 8)
        """
        super().__init__()
        self.num_classes = num_classes
        self.input_shape = input_shape

        self.cnn = nn.Sequential(
            Conv2dTF(in_channels=self.input_shape[0],
                     out_channels=conv1_filters,
                     kernel_size=conv1_kernel_size,
                     padding='same'
                     ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool1_size,
                         stride=pool1_stride),
            Conv2dTF(in_channels=conv1_filters,
                      out_channels=conv2_filters,
                      kernel_size=conv2_kernel_size,
                     padding = 'same'
                     ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool2_size,
                         stride=pool2_stride),
        )

        # determine number of features in output after stacking channels
        # we use the same number of features for hidden states
        # note self.num_hidden is also used to reshape output of cnn in self.forward method
        batch_shape = tuple((1,) + input_shape)
        tmp_tensor = torch.rand(batch_shape)
        tmp_out = self.cnn(tmp_tensor)
        channels_out, freqbins_out = tmp_out.shape[1], tmp_out.shape[2]
        self.num_rnn_features = channels_out * freqbins_out

        self.rnn = nn.LSTM(input_size=self.num_rnn_features,
                           hidden_size=self.num_rnn_features,
                           num_layers=1,
                           dropout=0,
                           bidirectional=True)

        # for self.fc, in_features = num_rnn_features * 2
        # because LSTM is bidirectional
        # so we get features forward + features backward as output
        self.fc = nn.Linear(self.num_rnn_features * 2, num_classes)

    def forward(self, x):
        features = self.cnn(x)
        # stack channels so that dimension order is (batch, num_rnn_features, num time bins)
        features = features.view(features.shape[0], self.num_rnn_features, -1)
        # now switch dimensions for feeding to rnn,
        # so dimension order is (num time bins, batch size, num_rnn_features)
        features = features.permute(2, 0, 1)
        rnn_output, (hidden, cell_state) = self.rnn(features)
        # permute back to (batch, time bins, features)
        # so we can project features down onto number of classes
        rnn_output = rnn_output.permute(1, 0, 2)
        logits = self.fc(rnn_output)
        # permute yet again
        # so that dimension order is (batch, classes, time steps)
        # because this is order that loss function expects
        return logits.permute(0, 2, 1)



"""functions to create plots, basis for figures in article"""
import matplotlib.pyplot as plt
import seaborn as sns


def frame_error(df, ax=None, figsize=(10, 3.75), legend=True, xlabel=False,
                save_as=None, transparent=True):
    """plot frame error results from generating a learning curve

    Parameters
    ----------
    df : pandas.DataFrame
        returned by article.util.make_df_birdsong_rec.
    figsize : tuple
        size of figure: (width, height) in inches.
        Default is (10, 3.75).
    xlabel : bool
        if True, add "training set size" label to x axis.
        Default is False (used when this function is called
        by error_rate_test_mean function).
    legend : bool
        if True, add legend to axis 0 (plot of individuals).
        Default is True.
    ax : matplotlib.axes.Axes
        axes on which to plot. Must be an array of size (1, 2).
        Default is None, in which case a new figure is created
        of size figsize, with the correct number of axes.
    save_as : str
        filename to save figure as
    transparent : bool
        if True, save figure as transparent. Default is True.

    Returns
    -------
    ax : matplotlib.axes.Axes
        either same as axes passed by user, or new axes created
        (when axes is None).
    """
    if ax is None:
        fig, ax = plt.subplots(1, 2, constrained_layout=True)
        fig.set_size_inches(figsize)

    df['frame_error_test_mean_pct'] = df['frame_error_test_mean'] * 100
    sns.pointplot(x="train_set_dur", y="frame_error_test_mean_pct",
                  hue="animal_ID", data=df,
                  linestyles=['--' for _ in df.animal_ID.unique()],
                  scale=0.75,
                  ax=ax[0])
    if xlabel:
        ax[0].set_xlabel("training set duration (s)", fontsize=20)
    else:
        ax[0].set_xlabel("")
    ax[0].set_ylabel("frame error (%)", fontsize=20)
    ax[0].set_ylim([0., 20.])
    ax[0].tick_params(labelsize=16)
    ax[0].legend(bbox_to_anchor=(1.01, 1), loc=2,
                 borderaxespad=0.)

    sns.boxplot(x="train_set_dur", y="frame_error_test_mean_pct", data=df, ax=ax[1])
    if xlabel:
        ax[1].set_xlabel("training set duration (s)", fontsize=20)
    else:
        ax[1].set_xlabel("")
    ax[1].set_ylabel("")
    ax[1].tick_params(labelsize=16)
    ax[1].set_ylim([0., 20.])

    if save_as:
        plt.savefig(save_as, bbox_inches='tight', transparent=transparent)

    return ax


def syllable_error_rate(df, ax=None, figsize=(10, 3.75), legend=False,
                        save_as=None, transparent=True):
    """plot syllable error rate results from generating a learning curve

    Parameters
    ----------
    df : pandas.DataFrame
        returned by article.util.make_df_birdsong_rec.
    figsize : tuple
        size of figure: (width, height) in inches.
        Default is (10, 3.75).
    legend : bool
        if True, add legend to axis 0 (plot of individuals).
        Default is True.
    ax : matplotlib.axes.Axes
        axes on which to plot. Must be an array of size (1, 2).
        Default is None, in which case a new figure is created
        of size figsize, with the correct number of axes.
    save_as : str
        filename to save figure as
    transparent : bool
        if True, save figure as transparent. Default is True.

    Returns
    -------
    ax : matplotlib.axes.Axes
        either same as axes passed by user, or new axes created
        (when axes is None).
    """
    if ax is None:
        fig, ax = plt.subplots(1, 2, constrained_layout=True)
        fig.set_size_inches(figsize)

    sns.pointplot(x="train_set_dur", y="syllable_error_test_mean",
                  hue="animal_ID", data=df,
                  linestyles=['--' for _ in df.animal_ID.unique()],
                  ax=ax[0])
    ax[0].set_xlabel("training set duration (s)", fontsize=20)
    ax[0].set_ylabel("syllable error rate\n(distance)", fontsize=20)
    ax[0].set_ylim([0., 1.5])
    ax[0].tick_params(labelsize=16)
    if legend:
        ax[0].legend(bbox_to_anchor=(1.01, 1), loc=2,
                     borderaxespad=0.)
    else:
        ax[0].legend_.remove()

    sns.boxplot(x="train_set_dur", y="syllable_error_test_mean", data=df, ax=ax[1])
    ax[1].set_xlabel("training set duration (s)", fontsize=20)
    ax[1].set_ylabel("")
    ax[1].set_ylim([0., 1.5])
    ax[1].tick_params(labelsize=16)
    label = 'Koumura\nOkanoya 2016'
    handle = ax[1].scatter([1, 2], [0.84, 0.46], s=75, color='k', marker='s',
                           label=label)
    ax[1].legend([handle], [label])

    if save_as:
        plt.savefig(save_as, bbox_inches='tight', transparent=transparent)

    return ax


def error_rate_test_mean(df, ax=None, figsize=(10, 7.5),
                         save_as=None, transparent=True):
    """plot both frame error and syllable error rate results from
    generating a learning curve.

    This function uses both functions defined above
    to reproduce the plots for the figure in the paper.

    Parameters
    ----------
    df : pandas.DataFrame
        returned by article.util.make_df_birdsong_rec.
    figsize : tuple
        size of figure: (width, height) in inches.
        Default is (10, 7.5).
    ax : matplotlib.axes.Axes
        axes on which to plot. Must be an array of size (2, 2).
        Default is None, in which case a new figure is created
        of size figsize, with the correct number of axes.
    save_as : str
        filename to save figure as
    transparent : bool
        if True, save figure as transparent. Default is True.

    Returns
    -------
    ax : matplotlib.axes.Axes
        either same as axes passed by user, or new axes created
        (when axes is None).s
    """
    if ax is None:
        fig, ax = plt.subplots(2, 2, constrained_layout=True)
        fig.set_size_inches(figsize)

    frame_error(df, ax[0, :], legend=True)
    syllable_error_rate(df, ax[1, :], legend=False)

    if save_as:
        plt.savefig(save_as, bbox_inches='tight', transparent=transparent)

    return ax



import numpy as np
import torch
import vak

from .network import TweetyNet


def acc(y_pred, y):
    return y_pred.eq(y.view_as(y_pred)).sum().item() / np.prod(y.shape)


class TweetyNetModel(vak.Model):
    @classmethod
    def from_config(cls, config, logger=None):
        network = TweetyNet(**config['network'])
        loss = torch.nn.CrossEntropyLoss(**config['loss'])
        optimizer = torch.optim.Adam(params=network.parameters(), **config['optimizer'])
        metrics = {'acc': vak.metrics.Accuracy(),
                   'loss': torch.nn.CrossEntropyLoss()}
        return cls(network=network, optimizer=optimizer, loss=loss, metrics=metrics, logger=logger)



"""functions used for syntax analysis"""
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd


HERE = Path(__file__).parent


def date_from_cbin_filename(cbin_filename):
    """extracts date as a string from a .cbin filename
    and converts to a datetime object

    Parameters
    ----------
    cbin_filename : str
        path to a .cbin file

    Returns
    -------
    dt : datetime
        convert from string using strptime function

    Examples
    --------
    >>> cbin_filename = 'gy6or6_baseline_220312_0836.3.cbin'
    >>> dt = date_from_cbin_filename(cbin_filename)
    """
    name = Path(cbin_filename).name
    splitname = name.split('.')  # splits into: stem, file number, extension
    splitname = splitname[0]  # just keep the stem
    splitname = splitname.split('_')  # further split stem; last two will be date + time
    date, time = splitname[-2:]
    dt = date + '-' + time
    dt = datetime.strptime(dt, '%d%m%y-%H%M')
    return dt


FIELDS_SYNTAX = [
    'date',
    'time',
    'cbin_filename',
    'label',
    'label_plus_one',
    'label_minus_one',
    'label_minus_two',
]


def make_df_trans_probs(vds_list, start_label='S', end_label='E'):
    """make a dataframe where each row represents a transition
    from one label to another, e.g. in annotated birdsong.

    Can be used to compute transition probabilities between birdsong syllables.

    Takes a list of vak Datasets; as written now, the function expects each
    vocalization in the dataset to be from .cbin audio files.


    Parameters
    ----------
    vds_list : list
        of vak.Dataset.
    start_label : str
        Label that represents start of a sequence, e.g. a song bout. Default is 'S'.
        The function adds a transition from this start state to the first labeled
        segment for every sequence, i.e. every vocalization in a dataset.
    end_label : str
        Label that represents end of a sequence. Default is 'E'.
        The function adds a transition from the first labeled segemtn to this end state
        for every sequence, i.e. every vocalization in a dataset.

    Returns
    -------
    df : pandas.Dataframe
        where each row represents one transition from one label to another
        and has the following columns:
            date : datetime
                extracted from audio filename.
                Used to determine transition probabilities on a certain day,
                and for resampling analysis where the dates are shuffled.
            cbin_filename : str
                filename of audio file associated with annotation.
            label : str
                String label for current label of interest.
                For example, label 'a' in transition 'a' -> 'b'.
            label_plus_one : str
                Label that follows current label.
                For exapmle, label 'b' in transition 'a' -> 'b'.
            label_minus_one : str
                Label that preceded current label.
            label_minus_two : str
                Label that preceded label before current label.
    """
    df_dict = {field: [] for field in FIELDS_SYNTAX}
    for vds in vds_list:
        for voc in vds.voc_list:
            cbin_filename = voc.audio_path
            datetime = date_from_cbin_filename(cbin_filename)
            labels = voc.annot.labels.tolist()
            labels = [start_label] + labels + [end_label]
            len_labels = len(labels)
            for ind, label in enumerate(labels):
                if ind == 0:
                    label_minus_one = None
                    label_minus_two = None
                    label_plus_one = labels[ind + 1]
                elif ind == 1:
                    label_minus_one = labels[0]
                    label_minus_two = None
                    label_plus_one = labels[ind + 1]
                elif ind == (len_labels - 1):  # i.e., last index
                    label_minus_one = labels[ind - 1]
                    label_minus_two = labels[ind - 2]
                    label_plus_one = None
                else:
                    label_minus_one = labels[ind - 1]
                    label_minus_two = labels[ind - 2]
                    label_plus_one = labels[ind + 1]
                df_dict['date'].append(datetime.date())
                df_dict['time'].append(datetime.time())
                df_dict['cbin_filename'].append(cbin_filename)
                df_dict['label'].append(label)
                df_dict['label_plus_one'].append(label_plus_one)
                df_dict['label_minus_one'].append(label_minus_one)
                df_dict['label_minus_two'].append(label_minus_two)
    df = pd.DataFrame.from_dict(df_dict)
    return df


def get_trans_prob(df, date, label, label_plus_one):
    """get probability of transition from one label to another,
    given a dataframe where rows are transitions

    Parameters
    ----------
    df : pandas.Dataframe
        returned by the make_df_trans_probs function
    date : datetime.date
        date to use to create the transition matrix.
        Must occur in the 'date' column of the dataframe
    label : str
        we want to compute the probability of
        transitioning from this label to label_plus_one
    label_plus_one : str
        label we want to compute the probability of
        transitioning to, from label_plus_one

    Returns
    -------
    p : float
        probability of transitioning from label to label_plus_one on
        the specified date, computed using the dataframe
    """
    df_date = df[df['date'] == date]
    label_count = len(
        df_date[df_date['label'] == label].index
    )
    trans_count = len(
        df_date[(df_date['label'] == label) & (df_date['label_plus_one'] == label_plus_one)].index
    )
    if label_count > 0:
        p = trans_count / label_count
    else:
        p = 0.
    return p


def make_trans_mat(df, date, min_p=0.01):
    """make transition matrix from a dataframe
    (returned by the make_df_trans_probs function)

    Parameters
    ----------
    df : pandas.Dataframe
        returned by the make_df_trans_probs function
    date : datetime.date
        date to use to create the transition matrix.
        Must occur in the 'date' column of the dataframe
    min_p : float
        minimum transition probability to keep.
        Default is 0.01. Any values below that will be
        set to 0.0, and the rows will be normalized so
        that they sum to 1.0.

    Returns
    -------
    trans_mat : numpy.ndarray
        transition matrix

    Notes
    -----
    Each row corresponds to one label from the unique set found
    in the dataframe, and the values in that row represent the
    probability of transitioning from that label to the label
    that corresponds to the column in which the value is found.
    """
    labels = df['label'].unique()
    num_labels = labels.shape[0]
    trans_mat = np.zeros((num_labels, num_labels))
    for row, label in enumerate(labels):
        if label == 'E':
            continue
        else:
            for col, label_plus_one in enumerate(labels):
                p = get_trans_prob(df, date, label, label_plus_one)
                if p > min_p:
                    trans_mat[row, col] = p
                else:
                    trans_mat[row, col] = 0.
    # adjust so all rows sum to 1
    row_sums = trans_mat.sum(axis=1)
    trans_mat = trans_mat[row_sums != 0.0, :]
    row_sums = trans_mat.sum(axis=1)
    trans_mat = trans_mat / row_sums[:, np.newaxis]
    return trans_mat


def find_branch_points(trans_mat, labels):
    """given a 2-d transition matrix, find branch points, i.e. labels
    that can transition to more than one label

    Parameters
    ----------
    trans_mat : numpy.ndarray
        2-d matrix where rows represent probabilities of transitioning
        from some label to whichever label is indexed by the column
        where the probability appears.
    labels : list
        of str, unique set of labels. The index of each label
        must correspond to the row in trans_mat representing the
        probabilities of transitioning from that label to others.

    Returns
    -------
    branch_point_ind : numpy.ndarray
        indices of rows that are branch points in trans_mat.
    branch_point_lbl : list
        of labels. Labels for indices in branch_point_ind.
    """
    branch_point_ind = []
    branch_point_lbl = []
    for ind, (row, label) in enumerate(zip(trans_mat, labels)):
        ps = np.nonzero(row)[0]
        if ps.shape[0] > 1:
            branch_point_ind.append(ind)
            branch_point_lbl.append(label)
    if branch_point_ind:  # is not empty
        branch_point_ind = np.asarray(branch_point_ind)
    return branch_point_ind, branch_point_lbl





"""utility functions used to munge data for article"""
from datetime import datetime
import json
from pathlib import Path
import pickle
from configparser import ConfigParser

import joblib
import pandas as pd

import vak


HERE = Path(__file__).parent


FIELDS_BIRDSONG_REC = [
    'animal_ID',
    'train_set_dur',
    'replicate',
    'net_name',
    'learning_rate',
    'time_bins',
    'frame_error_train',
    'frame_error_test',
    'syllable_error_train',
    'syllable_error_test',
]


def make_df_birdsong_rec(config_files, test_dirs, net_name='TweetyNet', csv_fname=None,
                         train_set_durs=None):
    """make Pandas dataframe of results from running vak.core.learning_curve,
    given list of config.ini files and lists of directories with results from
    vak.core.learning_curve.test

    Parameters
    ----------
    config_files : list
        of pathlib.Path objects; paths to config.ini files used to run vak.core.learning_curve
    test_dirs : list
        of pathlib.Path objects; directories output by vak.core.learning_curve.test.
        The parent directory of the file 'y_pres_and_err_for_train_and_test' is assumed to be
        the ID of the bird from which results were generated.
    net_name : str
        name of network that is also the name of the section in the config.ini file
        that defines its hyperparameters. Default is 'TweetyNet'.
    csv_fname : str
        filename used to save dataframe as a csv. Default is None, in which case no .csv is
        saved.
    train_set_durs : list
        of float, durations of training sets used, in seconds. Used to determine which
        accuracies from the test directory correspond to which training set duration.
        Default is None, in which case the values specified by the option 'train_set_durs'
        in the config.ini file is used. Those values may be different from the values used
        when running test.

    Returns
    -------
    results_df : Pandas.Dataframe
        with the following columns:
            animal_ID : str
                identity of animal from which vocalizations were recorded. Assumed to be name of test_dir that contains
                results. e.g. "data/BirdsongRecognition/Bird1/" would result in Bird1 being the animal_ID.
            train_set_dur : float
                duration of training data set. Determined from config.ini file.
            replicate : int
                training replicate, e.g. net number 1 of 5 trained with random subset
                with total duration `train_set_dur` from a larger set of training data.
                Determined from train_err file.
            learning_rate : float
                hyperparameter, update rate for weights in neural network. Determined
                from config.ini file.
            time_bins : int
                hyperparameter, number of time bins from spectrogram in windows shown to
                network. Determined from config.ini file.
            frame_error : float
                frame error rate, i.e. number of frames / time bins incorrectly classified
            syllable_error : float
                syllable error rate, i.e. Levenstein distance between original sequence of labels and sequence
                of labels recovered from predicted
    Notes
    -----
    This function assumes that config_files and test_dirs are in the same order, i.e. that
    config_files[0] is the config.ini file that produced the results in test_dirs[0].
    You can ensure this is the case by applying the `sorted` function to the lists, if
    you used the same name (e.g. "Bird1") in both the config.ini file and the test directory
    (or one of its parent directories, e.g. "Bird1/learning_curve/test/").
    """
    if train_set_durs is not None:
        if type(train_set_durs) != list:
            raise TypeError(
                f"train_set_durs should be a list but was {type(train_set_durs)}"
            )
        try:
            train_set_durs = [float(dur) for dur in train_set_durs]
        except ValueError:
            raise ValueError(
                'could not convert all elements in train_set_durs to float'
            )
    df_dict = {field: [] for field in FIELDS_BIRDSONG_REC}

    for config_file, test_dir in zip(config_files, test_dirs):
        # ------------ get what we need from config.ini ----------------------------------------------------------------
        config = ConfigParser()
        config.read(config_file)

        # ------------ need train set durations to pair with error values ----------------------------------------------
        if train_set_durs is None:
            train_set_durs = [float(element)
                              for element in
                              config['TRAIN']['train_set_durs'].split(',')]

        # ------------ then need hyperparameters -----------------------------------------------------------------------
        time_bins = config[net_name]['time_bins']
        learning_rate = config[net_name]['learning_rate']

        # ------------ now unpack everything from test_dir -------------------------------------------------------------
        animal_ID = test_dir.stem  # assume stem (i.e. last part of directory) is animal ID #
        train_err = list(test_dir.glob('train_err'))
        assert len(train_err) == 1, f"did not find only one train_err file: {train_err}"
        train_err = train_err[0]
        with open(train_err, 'rb') as f:
            frame_err_train = pickle.load(f)

        test_err = list(test_dir.glob('test_err'))
        assert len(test_err) == 1, f"did not find only one test_err file: {test_err}"
        test_err = test_err[0]
        with open(test_err, 'rb') as f:
            frame_err_test = pickle.load(f)

        preds_and_err = list(test_dir.glob('y_preds_and_err_for_train_and_test'))
        assert len(preds_and_err) == 1, f"did not find only one preds_and_err file: {preds_and_err}"
        preds_and_err = preds_and_err[0]
        preds_and_err = joblib.load(preds_and_err)
        syl_err_train = preds_and_err['train_syl_err_rate']
        syl_err_test = preds_and_err['test_syl_err_rate']

        err_zip = zip(train_set_durs, frame_err_train, frame_err_test, syl_err_train, syl_err_test)
        for train_set_dur, fe_tr_row, fe_te_row, se_tr_row, se_test_row in err_zip:
            err_rows = [fe_tr_row, fe_te_row, se_tr_row, se_test_row]
            row_lengths = [len(row) for row in err_rows]
            assert len(set(row_lengths)) == 1, f"found different lengths of rows in error arrays: {row_lengths}"
            for replicate, (fe_tr, fe_te, se_tr, se_te) in enumerate(zip(*err_rows)):
                df_dict['animal_ID'].append(animal_ID)
                df_dict['train_set_dur'].append(train_set_dur)
                df_dict['replicate'].append(replicate)
                df_dict['net_name'].append(net_name)
                df_dict['time_bins'].append(time_bins)
                df_dict['learning_rate'].append(learning_rate)
                df_dict['frame_error_train'].append(fe_tr)
                df_dict['frame_error_test'].append(fe_te)
                df_dict['syllable_error_train'].append(se_tr)
                df_dict['syllable_error_test'].append(se_te)

    df = pd.DataFrame.from_dict(df_dict)

    if csv_fname:
        df.to_csv(csv_fname)

    return df


def agg_df_birdsong_rec(df, train_set_durs):
    """filters dataframe of results by specified training set durations, then
    groups by animal ID and training set duration, and for each group computes
    the mean and standard deviation, i.e. aggregate statistics.

    The index of each row becomes one animal ID, and this function makes mean
    and standard deviations into columns, as well as the training set duration.
    This way the dataframe is in "long" form (where "animal ID" and "training set
    durations" are the conditions) for use with Seaborn.

    Parameters
    ----------
    df : pandas.Dataframe
    train_set_durs :

    Returns
    -------
    df_agg : pandas.Dataframe
    """
    df = df[df.train_set_dur.isin(train_set_durs)]
    df_agg = df.groupby(['animal_ID', 'train_set_dur']).agg({'frame_error_train': ['mean', 'std'],
                                                             'frame_error_test': ['mean', 'std'],
                                                             'syllable_error_train': ['mean', 'std'],
                                                             'syllable_error_test': ['mean', 'std']})
    df_agg.columns = ['_'.join(col).strip() for col in df_agg.columns.values]
    df_agg = df_agg.reset_index()
    return df_agg


# for Pandas dataframe / csv
FIELDS_BF_SONG_REPOSITORY = [
    'animal_ID',
    'test_dir_date',
    'test_dir_daynum',
    'test_set_dur',
    'train_set_dur',
    'net_name',
    'frame_error_train',
    'frame_error_test',
    'syllable_error_train',
    'syllable_error_test',
]


DEFAULT_DATA_DIR = HERE.joinpath('../../data/BFSongRepository')
TRAIN_SET_DUR = 60


def make_df_bf_song_repository(data_dir=DEFAULT_DATA_DIR, train_set_dur=TRAIN_SET_DUR,
                               net_name='TweetyNet', csv_fname=None):
    """make Pandas dataframe of results from running src/scripts/bfsongrepo-test-predict.py

    Parameters
    ----------
    data_dir : str
        path to directory with results from running bfsonrepo-test-predict.py.
        Default is './data/BirdsongRecognition' (relative to TweetyNet repository root).
    csv_fname : str
        filename used to save dataframe as a csv. Default is None, in which case no .csv is
        saved.

    Returns
    -------
    results_df : Pandas.Dataframe
        with the following columns:
            animal_ID : str
                identity of animal from which vocalizations were recorded. Assumed to be name of test_dir that contains
                results. e.g. "data/BirdsongRecognition/Bird1/" would result in Bird1 being the animal_ID.
            train_set_dur : float
                duration of training data set. Determined from config.ini file.
            replicate : int
                training replicate, e.g. net number 1 of 5 trained with random subset
                with total duration `train_set_dur` from a larger set of training data.
                Determined from train_err file.
            learning_rate : float
                hyperparameter, update rate for weights in neural network. Determined
                from config.ini file.
            time_bins : int
                hyperparameter, number of time bins from spectrogram in windows shown to
                network. Determined from config.ini file.
            frame_error : float
                frame error rate, i.e. number of frames / time bins incorrectly classified
            syllable_error : float
                syllable error rate, i.e. Levenstein distance between original sequence of labels and sequence
                of labels recovered from predicted
    Notes
    -----
    This function assumes that config_files and test_dirs are in the same order, i.e. that
    config_files[0] is the config.ini file that produced the results in test_dirs[0].
    You can ensure this is the case by applying the `sorted` function to the lists, if
    you used the same name (e.g. "Bird1") in both the config.ini file and the test directory
    (or one of its parent directories, e.g. "Bird1/learning_curve/test/").
    """
    data_dir = Path(data_dir)
    # need to add sub-directory that is duration of training set to data_dir
    data_dir = data_dir.joinpath(f'{train_set_dur}s')
    bird_subdirs = [subdir for subdir in data_dir.iterdir() if subdir.is_dir()]

    df_dict = {field: [] for field in FIELDS_BF_SONG_REPOSITORY}

    for bird_subdir in bird_subdirs:
        animal_ID = bird_subdir.name

        test_dataset_jsons = bird_subdir.joinpath('vds').glob('*.has_notmat.test.vds.json')
        test_dataset_jsons = sorted(test_dataset_jsons)

        test_results_jsons = bird_subdir.joinpath('json').glob('*.has_notmat.test.json')
        test_results_jsons = sorted(test_results_jsons)
        for json_ind, test_results_json in enumerate(test_results_jsons):
            json_results_name = test_results_json.name
            date_from_json_name = json_results_name.split('.')[0]

            test_dataset_json = [test_dataset_json
                                 for test_dataset_json in test_dataset_jsons
                                 if date_from_json_name in str(test_dataset_json)]
            if len(test_dataset_json) != 1:
                raise ValueError(
                    'did not find single test dataset matching results .json file,'
                    f'instead found: {test_dataset_json}'
                )
            else:
                test_dataset_json = test_dataset_json[0]
            test_vds = vak.Dataset.load(test_dataset_json)
            test_set_dur = sum([voc.duration for voc in test_vds.voc_list])

            test_dir_date = datetime.strptime(date_from_json_name, '%m%d%y')
            test_dir_daynum = json_ind + 1  # start numbering of days at 1
            with open(test_results_json) as fp:
                err_dict = json.load(fp)

            df_dict['animal_ID'].append(animal_ID)
            df_dict['test_dir_date'].append(test_dir_date)
            df_dict['test_dir_daynum'].append(test_dir_daynum)
            df_dict['train_set_dur'].append(train_set_dur)
            df_dict['test_set_dur'].append(test_set_dur)
            df_dict['net_name'].append(net_name)
            df_dict['frame_error_train'].append(err_dict['train_err'])
            df_dict['frame_error_test'].append(err_dict['test_err'])
            df_dict['syllable_error_train'].append(err_dict['train_syl_err_rate'])
            df_dict['syllable_error_test'].append(err_dict['test_syl_err_rate'])

    df = pd.DataFrame.from_dict(df_dict)

    if csv_fname:
        df.to_csv(csv_fname)

    return df
